@@ -77,6 +77,7 @@ Compile Level
 #include "ratcom.h"
 #include "lookup.h"
 #include "config.h"
+#include "getopt.h"
 
 #ifndef VERSION
 #define VERSION "(undefined)"
@@ -167,17 +168,19 @@ char deftyp[2] = {
 
 /* constant strings */
 
-char *errmsg = "error at line ";
-char *in     = " in ";
-char *ifnot  = "if(.not.";
-char *incl   = "include";
-char *fncn   = "function";
-char *def    = "define";
-char *bdef   = "DEFINE";
-char *contin = "continue";
-char *rgoto  = "goto ";
-char *dat    = "data ";
-char *eoss   = "EOS/";
+const char *errmsg = "error at line ";
+const char *in     = " in ";
+const char *ifnot  = "if(.not.";
+const char *incl   = "include";
+const char *fncn   = "function";
+const char *def    = "define";
+const char *bdef   = "DEFINE";
+const char *contin = "continue";
+const char *rgoto  = "goto ";
+const char *dat    = "data ";
+const char *eoss   = "EOS/";
+
+char stelem[128] = "integer";	/* element type of arrays generated for string directive */
 
 extern char ngetch(char *c, FILE *fd);
 char *progname;
@@ -185,36 +188,38 @@ int startlab = 23000;		/* default start label */
 int leaveC = NO;		/* Flag for handling comments */
 
 /* prototypes */
-extern int our_getopt(int argc, char **argv, char *optstring);
-extern struct hashlist *install(char *name, char *def);
-
-int ctoi(char *in, int *i);
-int equal(char *str1, char *str2);
-int gettok(char *token, int toksiz);
-int gnbtok(char *token, int toksiz);
-int gtok(char *lexstr, int toksiz, FILE *fd);
-int itoc(int n, char *str, int size);
-int labgen(int n);
-int lex(char *lexstr);
-int look(char *name,char *defn);
-int relate(char *token, FILE *fd);
-int type(char c);
-void baderr(char *msg);
+int alldig(const char *str);
+void baderr(const char *msg);
+void balpar(void);
 void brknxt(int sp, int *lextyp, int *labval, int token);
 void cascod (int lab, int token);
 int caslab (int *n, int *t);
+int ctoi(char *in_, int *i);
+int deftok(char *token, int toksiz, FILE *fd);
 void docode(int *lab);
+void dostat(int lab);
+void eatup(void);
 void elseif(int lab);
-void error(char *msg, char *s);
+int equal(const char *str1, const char *str2);
+void error(const char *msg, const char *s);
 void fold(char *token);
 void forcod(int *lab);
+void fors(int lab);
 void getdef(char *token, int toksiz, char *defn, int defsiz, FILE *fd);
+int gettok(char *token, int toksiz);
+int gnbtok(char *token, int toksiz);
+int gtok(char *lexstr, int toksiz, FILE *fd);
 void ifcode(int *lab);
-void ifthen(void);
+void ifend(void);
 void ifgo(int lab);
+void ifthen(void);
 void initvars(void);
-void labelc(char *lexstr);
-void otherc(char *lexstr);
+int itoc(int n, char *str, int size);
+void labelc(const char *lexstr);
+int labgen(int n);
+int lex(char *lexstr);
+int look(const char *name,char *defn);
+void otherc(const char *lexstr);
 void outasis(FILE *fd);
 void outch(char c);
 void outcmnt(FILE *fd);
@@ -222,21 +227,23 @@ void outcon(int n);
 void outdon(void);
 void outgo(int n);
 void outnum(int n);
-void outstr(char *str);
+void outstr(const char *str);
 void outtab(void);
 void parse(void);
-void pbstr(char *in);
+void pbstr(const char *string);
 void putbak(char c);
+int relate(char *token, FILE *fd);
 void repcod(int *lab);
 void retcod(void);
-void scopy(char *from, int i, char *to, int j);
+void scopy(const char *from, int i, char *to, int j);
 void skpblk(FILE *fd);
 void strdcl(void);
 void swcode (int *lab);
 void swend(int lab);
 void swvar (int lab);
-void synerr(char *msg);
-void unstak(int *sp, int *lextyp, int *labval, char token);
+void synerr(const char *msg);
+int type(signed char c);
+void unstak(int *sp, const int *lextyp, const int *labval, char token);
 void untils(int lab, int token);
 void whilec(int *lab);
 void whiles(int lab);
@@ -248,13 +255,11 @@ void whiles(int lab);
 int main(int argc,char **argv)
 {
 	int c, errflg = 0;
-	extern int optind77;
-	extern char *optarg;
 
 	progname = argv[0];
 
 /* ECL	while ((c=our_getopt(argc, argv, "Chn:l:o:6:")) != EOF) */
-	while ((c=our_getopt(argc, argv, "C?n:l:o:")) != EOF)
+	while ((c=our_getopt(argc, argv, "C?n:l:o:s:")) != EOF)
 	switch (c) {
 		case 'C':
 			leaveC = YES; /* keep comments in src */
@@ -275,6 +280,18 @@ int main(int argc,char **argv)
 			if ((freopen(optarg, "w", stdout)) == NULL)
 				error("can't write %s\n", optarg);
 			break;
+		case 's':
+			if (optarg != NULL) {
+			        if (strlen(optarg) > sizeof(stelem) - 1) {
+				        fprintf(stderr, "-s argument too long\n");
+					exit(1);
+				}
+			        strncpy(stelem, optarg, sizeof(stelem) - 1);
+				stelem[sizeof(stelem) - 1] = '\0';
+			} else {
+				fprintf(stderr, "-s requires a type name argument\n");
+				exit(1);
+			}
 		case '6':
 				/* not written yet */
 			break;
@@ -422,7 +439,7 @@ void parse(void)
  *  alldig - return YES if str is all digits
  *
  */
-int alldig(char *str)
+int alldig(const char *str)
 {
 	int i,j;
 
@@ -476,8 +493,7 @@ void balpar(void)
  * deftok - get token; process macro calls and invocations
  *
  */
-int
-deftok(char *token, int toksiz, FILE *fd)
+int deftok(char *token, int toksiz, FILE *fd)
 {
 	char defn[MAXDEF];
 	int t;
@@ -499,14 +515,12 @@ deftok(char *token, int toksiz, FILE *fd)
 	return(t);
 }
 
-
 /*
  * eatup - process rest of statement; interpret continuations
  *
  */
 void eatup(void)
 {
-
 	char ptoken[MAXTOK], token[MAXTOK];
 	int nlpar, t;
 
@@ -697,6 +711,7 @@ int gnbtok(char *token, int toksiz)
 {
 	int tok;
 
+	if (level < 0) return EOF;
 	skpblk(infile[level]);
 	tok = gettok(token, toksiz);
 	return(tok);
@@ -715,20 +730,18 @@ int gtok(char *lexstr, int toksiz, FILE *fd)
 		lexstr[0] = BLANK;
 		while (c == BLANK || c == TAB)    /* compress many blanks to one */
 			c = ngetch(&c, fd);
-		if (c == PERCENT) 
-		{
+		if (c == PERCENT) {
 			  outasis(fd);		/* copy direct to output if % */
 			  c = NEWLINE;
 		}
 		if (c == SHARP) {
-			if(leaveC == YES)
-			{
-			  outcmnt(fd);		/* copy comments to output */
-			  c = NEWLINE;
+			if (leaveC == YES) {
+			        outcmnt(fd);		/* copy comments to output */
+			        c = NEWLINE;
+			} else {
+			        while (ngetch(&c, fd) != NEWLINE) /* strip comments */
+				        ;
 			}
-			else
-			  while (ngetch(&c, fd) != NEWLINE) /* strip comments */
-				;
 		}
 /*
 		if (c == UNDERLINE)	
@@ -822,12 +835,12 @@ int gtok(char *lexstr, int toksiz, FILE *fd)
 		tok = NEWLINE;
 	}
 	else if (c == SHARP) { 
-		if(leaveC == YES)
-		  outcmnt(fd);		/* copy comments to output */
+		if (leaveC == YES)
+		        outcmnt(fd);		/* copy comments to output */
 		else
-		  while (ngetch(&lexstr[0], fd) != NEWLINE) /* strip comments */
-			;
-		  tok = NEWLINE;
+		        while (ngetch(&lexstr[0], fd) != NEWLINE) /* strip comments */
+			        ;
+		tok = NEWLINE;
 	}
 	else if (c == GREATER || c == LESS || c == NOT
 		 || c == BANG || c == CARET || c == EQUALS
@@ -946,16 +959,17 @@ char ngetch(char *c, FILE *fd)
 
 	return(*c);
 }
+
 /*
  * pbstr - push string back onto input
  *
  */
-void pbstr(char *in)
+void pbstr(const char *string)
 {
 	int i;
 
-	for (i = strlen(in) - 1; i >= 0; i--)
-		putbak(in[i]);
+	for (i = strlen(string) - 1; i >= 0; i--)
+		putbak(string[i]);
 }
 
 /*
@@ -964,21 +978,18 @@ void pbstr(char *in)
  */
 void putbak(char c)
 {
-
 	bp++;
 	if (bp > BUFSIZE)
 		baderr("too many characters pushed back.");
 	buf[bp] = c;
 }
 
-
 /*
  * relate - convert relational shorthands into long form
  *
  */
 int relate(char *token, FILE *fd)
 {
-
 	if (ngetch(&token[1], fd) != EQUALS) {
 		putbak(token[1]);
 		token[2] = LETT;
@@ -1041,12 +1052,15 @@ void skpblk(FILE *fd)
 	putbak(c);
 }
 
-
 /*
  * type - return LETTER, DIGIT or char; works with ascii alphabet
  *
+ * to preserve the identity of EOF when casting back and forth between the
+ * parameter value of type() and int, we use 'signed char' explicitly here,
+ * so that the code also works on systems that use unsigned chars by default.
+ *
  */
-int type(char c)
+int type(signed char c)
 {
 	int t;
 
@@ -1313,9 +1327,8 @@ void ifthen(void)
  * labelc - output statement number
  *
  */
-void labelc(char *lexstr)
+void labelc(const char *lexstr)
 {
-
 	xfer = NO;   /* can't suppress goto's now */
 	if (strlen((char *) (&lexstr[0])) == 5)   /* warn about 23xxx labels */
 		if (atoi((char*)lexstr) >= startlab)
@@ -1341,7 +1354,7 @@ int labgen(int n)
  * otherc - output ordinary Fortran statement
  *
  */
-void otherc(char *lexstr)
+void otherc(const char *lexstr)
 {
 	xfer = NO;
 	outtab();
@@ -1391,7 +1404,6 @@ void outcon(int n)
  */
 void outdon(void)
 {
-
 	outbuf[outp] = NEWLINE;
 	outbuf[outp+1] = EOS;
 	printf("%s", outbuf);
@@ -1435,7 +1447,7 @@ void outasis(FILE *fd)
 {
 	char c;
 	while((c=ngetch(&c,fd)) != NEWLINE)
-					outch(c);
+		outch(c);
 	outdon();
 }
 
@@ -1459,7 +1471,6 @@ void outgo(int n)
  */
 void outnum(int n)
 {
-
 	char chars[MAXCHARS];
 	int i, m;
 
@@ -1477,13 +1488,11 @@ void outnum(int n)
 		outch(chars[i]);
 }
 
-
-
 /*
  * outstr - output string
  *
  */
-void outstr(char *str)
+void outstr(const char *str)
 {
 	int i;
 
@@ -1501,14 +1510,12 @@ void outtab(void)
 		outch(BLANK);
 }
 
-
 /*
  * repcod - generate code for beginning of repeat
  *
  */
 void repcod(int *lab)
 {
-
 	int tlab;
 
 	tlab = *lab;
@@ -1543,36 +1550,60 @@ void retcod(void)
 	xfer = YES;
 }
 
-
-/* strdcl - generate code for string declaration */
+/*
+ * strdcl - generate code for string declaration
+ *
+ */
 void strdcl(void)
 {
-	char t, name[MAXNAME], init[MAXTOK];
+	char t, name[MAXNAME], init[MAXTOK], size[MAXTOK];
 	int i, len;
+	int quoted = 0, actlen = 0;
 
 	t = gnbtok(name, MAXNAME);
 	if (t != ALPHA)
 		synerr("missing string name.");
 	if (gnbtok(init, MAXTOK) != LPAREN) {  /* make size same as initial value */
-		len = strlen((char *) (&init[0])) + 1;
-		if (init[1] == SQUOTE || init[1] == DQUOTE)
-			len = len - 2;
+		if (init[0] == SQUOTE || init[0] == DQUOTE) {
+			/* minus two for quotes; plus one for eos */
+			actlen = strlen(init) - 1;
+			quoted = 1;
+		} else {
+			/* plus one for eos */
+			actlen = strlen(init) + 1;
+		}
+		len = actlen;
+		/* gtok should match up the quotes for us. */
 	}
 	else {	/* form is string name(size) init */
-		t = gnbtok(init, MAXTOK);
+		t = gnbtok(size, MAXTOK);
 		i = 0;
-		len = ctoi(init, &i);
-		if (init[i] != EOS)
+		len = ctoi(size, &i);
+		if (size[i] != EOS)
 			synerr("invalid string size.");
+		/* we read the size, now consume the right parenthesis */
 		if (gnbtok(init, MAXTOK) != RPAREN)
 			synerr("missing right paren.");
-		else
+		else {
+			/* and finally read the initializer */
 			t = gnbtok(init, MAXTOK);
+		}
+		if (init[0] == SQUOTE || init[0] == DQUOTE) {
+			/* minus two for quotes; plus one for eos */
+			actlen = strlen(init) - 1;
+			quoted = 1;
+		} else {
+			/* plus one for eos */
+			actlen = strlen(init) + 1;
+		}
+		if (len < 1) {
+			synerr("specified string size too small.");
+			len = 1;
+		}
 	}
 	outtab();
-	/*
-	*   outstr(int);
-	*/
+	outstr(stelem);
+	outch(BLANK);
 	outstr(name);
 	outch(LPAREN);
 	outnum(len);
@@ -1580,11 +1611,16 @@ void strdcl(void)
 	outdon();
 	outtab();
 	outstr(dat);
-	len = strlen((char *)(&init[0])) + 1;
-	if (init[0] == SQUOTE || init[0] == DQUOTE) {
-		init[len-1] = EOS;
-		scopy(init, 1, init, 0);
-		len = len - 2;
+	if (quoted) {
+		init[actlen] = EOS; /* delete the ending quote */
+		scopy(init, 1, init, 0); /* delete the starting quote */
+	}
+	/* make sure user-provided size will result in a reasonable data stmt */
+	if (len > actlen) {
+		len = actlen; /* only initialize beginning of string */
+	}
+	else if (len < actlen) {
+		init[len-1] = EOS; /* truncate string to user-specified size */
 	}
 	for (i = 1; i <= len; i++) {	/* put out variable names */
 		outstr(name);
@@ -1595,7 +1631,6 @@ void strdcl(void)
 			outch(COMMA);
 		else
 			outch(SLASH);
-		;
 	}
 	for (i = 0; init[i] != EOS; i++) {	/* put out init */
 		outnum(init[i]);
@@ -1604,12 +1639,11 @@ void strdcl(void)
 	pbstr(eoss);	/* push back EOS for subsequent substitution */
 }
 
-
 /*
  * unstak - unstack at end of statement
  *
  */
-void unstak(int *sp, int *lextyp, int *labval, char token)
+void unstak(int *sp, const int *lextyp, const int *labval, char token)
 {
 	int tp;
 
@@ -1708,7 +1742,7 @@ void whiles(int lab)
 /*
  *  baderr - print error message, then die
  */
-void baderr(char *msg)
+void baderr(const char *msg)
 {
 	synerr(msg);
 	exit(1);
@@ -1717,16 +1751,16 @@ void baderr(char *msg)
 /*
  * error - print error message with one parameter, then die
  */
-void error(char *msg, char *s)
+void error(const char *msg, const char *s)
 {
-	fprintf(stderr, msg,s);
+	fprintf(stderr, msg, s);
 	exit(1);
 }
 
 /*
  * synerr - report Ratfor syntax error
  */
-void synerr(char *msg)
+void synerr(const char *msg)
 {
 	char lc[MAXCHARS];
 	int i;
@@ -1755,17 +1789,17 @@ void synerr(char *msg)
 /*
  * ctoi - convert string at in[i] to int, increment i
  */
-int ctoi(char *in, int *i)
+int ctoi(char *in_, int *i)
 {
 	int k, j;
 
 	j = *i;
-	while (in[j] == BLANK || in[j] == TAB)
+	while (in_[j] == BLANK || in_[j] == TAB)
 		j++;
-	for (k = 0; in[j] != EOS; j++) {
-		if (in[j] < DIG0 || in[j] > DIG9)
+	for (k = 0; in_[j] != EOS; j++) {
+		if (in_[j] < DIG0 || in_[j] > DIG9)
 			break;
-		k = 10 * k + in[j] - DIG0;
+		k = 10 * k + in_[j] - DIG0;
 	}
 	*i = j;
 	return(k);
@@ -1794,7 +1828,7 @@ void fold(char *token)
  * equal - compare str1 to str2; return YES if equal, NO if not
  *
  */
-int equal(char *str1, char *str2)
+int equal(const char *str1, const char *str2)
 {
 	int i;
 
@@ -1808,7 +1842,7 @@ int equal(char *str1, char *str2)
  * scopy - copy string at from[i] to to[j]
  *
  */
-void scopy(char *from, int i, char *to, int j)
+void scopy(const char *from, int i, char *to, int j)
 {
 	int k1, k2;
 
@@ -1824,9 +1858,8 @@ void scopy(char *from, int i, char *to, int j)
  * look - look-up a definition
  *
  */
-int look(char *name,char *defn)
+int look(const char *name,char *defn)
 {
-	extern struct hashlist *lookup();
 	struct hashlist *p;
 
 	if ((p = lookup(name)) == NULL)
@@ -2004,15 +2037,15 @@ void swend(int lab)
 	int lb, ub;
 	int j;
 
-static	char *slt   	= ".lt.1.or.";
-static	char *sgt   	= ".gt.";
-static	char *sgoto 	= "goto (";
+static	const char *slt   	= ".lt.1.or.";
+static	const char *sgt   	= ".gt.";
+static	const char *sgoto 	= "goto (";
 #endif
-static	char *sif   	= "if (";
-static	char *seq   	= ".eq.";
-static	char *sge   	= ".ge.";
-static	char *sle   	= ".le.";
-static	char *sand  	= ".and.";
+static	const char *sifp   	= "if (";
+static	const char *seq   	= ".eq.";
+static	const char *sge   	= ".ge.";
+static	const char *sle   	= ".le.";
+static	const char *sand  	= ".and.";
 
 #ifdef OUTPUT_BRANCH_TABLE
 	lb = swstak[swtop + 3];
@@ -2038,7 +2071,7 @@ static	char *sand  	= ".and.";
 			outdon();
 		}
 		outtab();  /*  if (Innn.lt.1.or.Innn.gt.ub-lb+1)goto default */
-		outstr (sif);
+		outstr (sifp);
 		swvar  (lab);
 		outstr (slt);
 		swvar  (lab);
@@ -2071,7 +2104,7 @@ static	char *sand  	= ".and.";
 	if (n > 0) { 		/* # output linear search form */
 		for (i = swtop + 3; i < swlast; i = i + 3) {
 			outtab();		/* # if (Innn */
-			outstr (sif);
+			outstr (sifp);
 			swvar  (lab);
 			if (swstak[i] == swstak[i+1]) {
 				outstr (seq); 	/* #   .eq....*/
@@ -2101,7 +2134,6 @@ static	char *sand  	= ".and.";
  */
 void swvar (int lab)
 {
-
 	outch ('I');
 	outnum (lab);
 }
